import copy
import logging

import pandas as pd
from pandas import DataFrame

from autogluon.core.features.types import R_INT, R_FLOAT, S_BINNED

from .abstract import AbstractFeatureGenerator
from .. import binning
from ..utils import get_smallest_valid_dtype_int

logger = logging.getLogger(__name__)


# TODO: Add more parameters (possibly pass in binning function as an argument for full control)
class BinnedFeatureGenerator(AbstractFeatureGenerator):
    """BinnedFeatureGenerator bins incoming int and float features to num_bins unique int values, maintaining relative rank order."""
    def __init__(self, num_bins=10, inplace=False, **kwargs):
        super().__init__(**kwargs)
        self.num_bins = num_bins
        self.inplace = inplace

    def _fit_transform(self, X: DataFrame, **kwargs) -> (DataFrame, dict):
        self._bin_map = self._get_bin_map(X=X)
        self._astype_map = {feature: get_smallest_valid_dtype_int(min_val=0, max_val=len(bin_index)) for feature, bin_index in self._bin_map.items()}
        X_out = self._transform(X)
        type_group_map_special = copy.deepcopy(self.feature_metadata_in.type_group_map_special)
        type_group_map_special[S_BINNED] += list(X_out.columns)
        return X_out, type_group_map_special

    def _transform(self, X: DataFrame) -> DataFrame:
        return self._transform_bin(X)

    @staticmethod
    def get_default_infer_features_in_args() -> dict:
        return dict(valid_raw_types=[R_INT, R_FLOAT])

    def _get_bin_map(self, X: DataFrame) -> dict:
        return binning.generate_bins(X, list(X.columns), ideal_bins=self.num_bins)

    def _transform_bin(self, X: DataFrame):
        if self._bin_map:
            if not self.inplace:
                X = X.copy(deep=True)
            with pd.option_context('mode.chained_assignment', None):
                # Pandas complains about SettingWithCopyWarning, but this should be valid.
                for column in self._bin_map:
                    X[column] = binning.bin_column(series=X[column], mapping=self._bin_map[column], dtype=self._astype_map[column])
        return X

    def _remove_features_in(self, features: list):
        super()._remove_features_in(features)
        if self._bin_map:
            for feature in features:
                if feature in self._bin_map:
                    self._bin_map.pop(feature)
        if self._astype_map:
            for feature in features:
                if feature in self._astype_map:
                    self._astype_map.pop(feature)

    def _more_tags(self):
        return {'feature_interactions': False}



import logging
from collections import defaultdict

import numpy as np
import pandas as pd
from pandas import DataFrame, IntervalIndex, Series

logger = logging.getLogger(__name__)


def bin_column(series: Series, mapping, dtype):
    mapping_dict = {k: v for v, k in enumerate(list(mapping))}
    series_out = pd.cut(series, mapping)
    series_out_int = np.array([mapping_dict[val] for val in series_out], dtype=dtype)
    return series_out_int


# TODO: Rewrite with normalized value counts as binning technique, will be more performant and optimal
def generate_bins(X_features: DataFrame, features_to_bin: list, ideal_bins: int = 10):
    X_len = len(X_features)
    starting_cats = 1000
    bin_index_starting = [np.floor(X_len * (num + 1) / starting_cats) for num in range(starting_cats - 1)]
    bin_epsilon = 0.000000001
    bin_mapping = defaultdict()
    max_iterations = 20
    for column in features_to_bin:
        num_cats_initial = starting_cats
        bins_value_counts = X_features[column].value_counts(ascending=False, normalize=True)
        max_bins = len(bins_value_counts)

        if max_bins <= ideal_bins:
            bins = pd.Series(data=sorted(X_features[column].unique()))
            num_cats_initial = max_bins
            cur_len = max_bins
            bin_index = list(range(num_cats_initial))
            interval_index = get_bins(bins=bins, bin_index=bin_index, bin_epsilon=bin_epsilon)
        else:
            cur_len = X_len
            bins = X_features[column].sort_values(ascending=True)
            interval_index = get_bins(bins=bins, bin_index=bin_index_starting, bin_epsilon=bin_epsilon)

        # TODO: max_desired_bins and min_desired_bins are currently equivalent, but in future they will be parameterized to allow for flexibility.
        max_desired_bins = min(ideal_bins, max_bins)
        min_desired_bins = min(ideal_bins, max_bins)

        is_satisfied = (len(interval_index) >= min_desired_bins) and (len(interval_index) <= max_desired_bins)

        num_cats_current = num_cats_initial
        cur_iteration = 0
        while not is_satisfied:
            ratio_reduction = max_desired_bins / len(interval_index)
            num_cats_current = int(np.floor(num_cats_current * ratio_reduction))
            bin_index = [np.floor(cur_len * (num + 1) / num_cats_current) for num in range(num_cats_current - 1)]
            interval_index = get_bins(bins=bins, bin_index=bin_index, bin_epsilon=bin_epsilon)

            if (len(interval_index) >= min_desired_bins) and (len(interval_index) <= max_desired_bins):
                is_satisfied = True
                # print('satisfied', column, len(interval_index))
            cur_iteration += 1
            if cur_iteration >= max_iterations:
                is_satisfied = True
                # print('max_iterations met, stopping prior to satisfaction!', column, len(interval_index))

        bin_mapping[column] = interval_index
    return bin_mapping


# TODO: Clean code
# TODO: Consider re-using bins variable instead of making bins_2-7 variables
# bins is a sorted int/float series, ascending=True
def get_bins(bins: Series, bin_index: list, bin_epsilon: float) -> IntervalIndex:
    max_val = bins.max()
    bins_2 = bins.iloc[bin_index]
    bins_3 = list(bins_2.values)
    bins_unique = sorted(list(set(bins_3)))
    bins_with_epsilon_max = set([i for i in bins_unique] + [i - bin_epsilon for i in bins_unique if i == max_val])
    removal_bins = set([bins_unique[index - 1] for index, i in enumerate(bins_unique[1:], start=1) if i == max_val])
    bins_4 = sorted(list(bins_with_epsilon_max - removal_bins))
    bins_5 = [np.inf if (x == max_val) else x for x in bins_4]
    bins_6 = sorted(list(set([-np.inf] + bins_5 + [np.inf])))
    bins_7 = [(bins_6[i], bins_6[i + 1]) for i in range(len(bins_6) - 1)]
    interval_index = IntervalIndex.from_tuples(bins_7)
    return interval_index
